#version 330 core

const int MAX_POINT_LIGHTS = 2; 

in vec2 fragmentUV;
in vec3 fragmentNormal;
in vec3 worldPos;
in vec4 lightSpacePos;

layout(location=0) out vec4 fragColor;

struct BaseLight                                                                    
{                                                                                   
    vec3 Color;                                                                     
    float AmbientIntensity;                                                         
    float DiffuseIntensity;                                                         
};                                                                                  
                                                                                    
struct DirectionalLight                                                             
{                                                                                   
    BaseLight Base;                                                                 
    vec3 Direction;                                                                 
};                                                                                  
                                                                                    
struct Attenuation                                                                  
{                                                                                   
    float Constant;                                                                 
    float Linear;                                                                   
    float Exp;                                                                      
};                                                                                  
                                                                                    
struct PointLight                                                                           
{                                                                                           
    BaseLight Base;                                                                         
    vec3 Position;                                                                          
    Attenuation Atten;                                                                      
};

struct SpotLight                                                                            
{                                                                                           
    PointLight Base;                                                                        
    vec3 Direction;                                                                         
    float Cutoff;                                                                           
};    

uniform sampler2D colorTexture;
uniform sampler2DShadow shadowMap;
uniform vec2 mapSize;
uniform DirectionalLight directionalLight;
uniform vec3 eyeWorldPos;
uniform float matSpecularIntensity;
uniform int numPointLights; 
uniform PointLight pointLights[MAX_POINT_LIGHTS];  
uniform SpotLight spotLight;
//Materials
uniform vec4 mDiffuse;
uniform vec4 mSpecular;
uniform vec4 mAmbient;
uniform float specularPower; 
//uniform vec4 mEmission;
#define EPSILON 0.00001

float CalcShadowFactor(vec4 LightSpacePos)
{
	vec3 ProjCoords = LightSpacePos.xyz / LightSpacePos.w;
    vec2 UVCoords;
    UVCoords.x = 0.5 * ProjCoords.x + 0.5;
    UVCoords.y = 0.5 * ProjCoords.y + 0.5;
    float z = 0.5 * ProjCoords.z + 0.5;

    float xOffset = 1.0/mapSize.x;
    float yOffset = 1.0/mapSize.y;

    float Factor = 0.0;

    for (int y = -1 ; y <= 1 ; y++) {
        for (int x = -1 ; x <= 1 ; x++) {
            vec2 Offsets = vec2(x * xOffset, y * yOffset);
            vec3 UVC = vec3(UVCoords + Offsets, z + EPSILON);
            Factor += texture(shadowMap, UVC);
        }
    }

    return (0.5 + (Factor / 18.0));
}

vec4 CalcLightInternal(BaseLight light, vec3 lightDirection, vec3 normal, float shadowFactor)                   
{                                                                                           
    vec4 ambientColor = mAmbient * vec4(light.Color, 1.0f) * light.AmbientIntensity;                   
    float diffuseFactor = dot(normal, -lightDirection);                                     
                                                                                            
    vec4 diffuseColor  = vec4(0, 0, 0, 0);                                                  
    vec4 specularColor = vec4(0, 0, 0, 0);                                                  
                                                                                            
    if (diffuseFactor > 0) {                                                                
        diffuseColor = mDiffuse * vec4(light.Color, 1.0f) * light.DiffuseIntensity * diffuseFactor;    
                                                                                            
        vec3 vertexToEye = normalize(eyeWorldPos - worldPos);                             
        vec3 lightReflect = normalize(reflect(lightDirection, normal));                     
        float specularFactor = dot(vertexToEye, lightReflect);                              
        specularFactor = pow(specularFactor, specularPower);                               
        if (specularFactor > 0) {                                                           
            specularColor =mSpecular * vec4(light.Color, 1.0f) *                                       
                            matSpecularIntensity * specularFactor;                         
        }                                                                                   
    }                                                                                       
                                                                                            
    return (ambientColor + (shadowFactor *(diffuseColor + specularColor)));                                   
}     

vec4 CalcDirectionalLight(vec3 normal)                                                      
{                                                                                           
    return CalcLightInternal(directionalLight.Base, directionalLight.Direction, normal, 1.0); 
}             

vec4 CalcPointLight(PointLight l, vec3 normal, vec4 LightSpacePos)                                                 
{                                                                                           
    vec3 lightDirection = worldPos - l.Position;                         
    float distance = length(lightDirection);                                                
    lightDirection = normalize(lightDirection);                                             
	float ShadowFactor = CalcShadowFactor(LightSpacePos);
						                                                                    
    vec4 color = CalcLightInternal(l.Base, lightDirection, normal, ShadowFactor);       
    float attenuation =  l.Atten.Constant +                               
                         l.Atten.Linear * distance +                      
                         l.Atten.Exp * distance * distance;               
                                                                                            
    return color / attenuation;                                                             
}              

vec4 CalcSpotLight(SpotLight l, vec3 normal, vec4 LightSpacePos)
{
    vec3 lightToPixel = normalize(worldPos - l.Base.Position);
    float spotFactor = dot(lightToPixel, l.Direction);

    if (spotFactor > l.Cutoff) {
        vec4 color = CalcPointLight(l.Base, normal, LightSpacePos);
        return color * (1.0 - (1.0 - spotFactor) * 1.0/(1.0 - l.Cutoff));
    }
    else {
        return vec4(0,0,0,0);
    }
}    



void main()
{
	vec3 normal = normalize(fragmentNormal);                                                       
    vec4 totalLight = CalcDirectionalLight(normal);                                         
                                                                                            
    for (int i = 0 ; i < numPointLights ; i++) {                                           
        totalLight += CalcPointLight(pointLights[i], normal, lightSpacePos);                                            
    }                                                                                       
    totalLight += CalcSpotLight(spotLight, normal, lightSpacePos);
	                                                                         
    fragColor = texture2D(colorTexture, fragmentUV.xy) * totalLight;


}